# Copyright 2021 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import operator
import os

from copy import deepcopy

cimport cython

from cython.operator cimport preincrement as inc, dereference as deref
from libc.math cimport ceil, floor
from libc.string cimport memcpy
from libcpp.vector cimport vector

import dimod

from dimod.binary.cybqm cimport cyBQM
from dimod.cyutilities cimport as_numpy_float, ConstInteger, ConstNumeric, cppvartype
from dimod.libcpp.vartypes cimport vartype_info as cppvartype_info
from dimod.quadratic cimport cyQM
from dimod.sampleset import as_samples
from dimod.variables import Variables
from dimod.vartypes import as_vartype, Vartype


ctypedef fused cyBQM_and_QM:
    cyBQM
    cyQM


cdef class cyQM_template(cyQMBase):
    def __cinit__(self):
        self.cppqm = self.base = new cppQuadraticModel[bias_type, index_type]()

    def __dealloc__(self):
        if self.cppqm is not NULL:
            del self.cppqm

    def __init__(self):
        super().__init__()
        self.REAL_INTERACTIONS = dimod.REAL_INTERACTIONS

    def __deepcopy__(self, memo):
        cdef cyQM_template new = type(self)()
        new.cppqm[0] = self.cppqm[0]  # *cppqm = *cppqm
        new.variables = deepcopy(self.variables, memo)

        new.REAL_INTERACTIONS = self.REAL_INTERACTIONS

        memo[id(self)] = new
        return new

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def _ilower_triangle_load(self, Py_ssize_t vi, Py_ssize_t num_neighbors, const unsigned char[:] buff):
        cdef Py_ssize_t index_itemsize = self.index_dtype.itemsize
        cdef Py_ssize_t bias_itemsize = self.dtype.itemsize
        cdef Py_ssize_t itemsize = index_itemsize + bias_itemsize

        if num_neighbors*itemsize > buff.size:
            raise RuntimeError

        cdef Py_ssize_t i
        cdef index_type ui
        cdef bias_type bias
        for i in range(num_neighbors):
            memcpy(&ui, &buff[i*itemsize], index_itemsize)
            memcpy(&bias, &buff[i*itemsize+index_itemsize], bias_itemsize)

            self.cppqm.add_quadratic_back(ui, vi, bias)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def _ivartypes_load(self, buff, Py_ssize_t num_variables):
        if self.num_variables():
            raise RuntimeError("cannot load vartypes into a model with variables")

        # use the bias size to determine the vartype size since we're not
        # letting numpy handle the alignment
        dtype = np.dtype([('vartype', np.int8),
                          ('lb', BIAS_DTYPE), ('ub', BIAS_DTYPE)],
                         align=False)

        arr = np.frombuffer(buff[:dtype.itemsize*num_variables], dtype=dtype)
        cdef const np.int8_t[:] vartype_view = arr['vartype']
        cdef const bias_type[:] lb_view = arr['lb']
        cdef const bias_type[:] ub_view = arr['ub']

        cdef Py_ssize_t vi
        cdef cppVartype cpp_vartype
        for vi in range(num_variables):
            # I hate that we have to do this manually, but cython doesn't
            # really like static_cast in this context
            if vartype_view[vi] == 0:
                cpp_vartype = cppVartype.BINARY
            elif vartype_view[vi] == 1:
                cpp_vartype = cppVartype.SPIN
            elif vartype_view[vi] == 2:
                cpp_vartype = cppVartype.INTEGER
            elif vartype_view[vi] == 3:
                cpp_vartype = cppVartype.REAL
            else:
                raise RuntimeError
            self.cppqm.add_variable(cpp_vartype, lb_view[vi], ub_view[vi])

        while self.variables.size() < self.cppqm.num_variables():
            self.variables._append()

    def add_linear(self, v, bias_type bias, *,
                   default_vartype=None,
                   default_lower_bound=None,
                   default_upper_bound=None,
                   ):
        cdef Py_ssize_t vi

        if default_vartype is None or self.variables.count(v):
            # already present
            vi = self.variables.index(v)
        else:
            # we need to add it
            vi = self.num_variables()
            self.add_variable(default_vartype, v,
                              lower_bound=default_lower_bound,
                              upper_bound=default_upper_bound,
                              )

        self.cppqm.add_linear(vi, bias)

    def add_linear_from_array(self, ConstNumeric[:] linear):
        cdef Py_ssize_t length = linear.shape[0]
        cdef Py_ssize_t vi

        if self.variables._is_range():
            if length > self.num_variables():
                raise ValueError("variables must already exist")
            for vi in range(length):
                self.cppqm.add_linear(vi, linear[vi])
        else:
            for vi in range(length):
                self.add_linear(vi, linear[vi])

    cdef Py_ssize_t _add_quadratic(self, index_type ui, index_type vi, bias_type bias) except -1:
        # note: does not test that ui, vi are valid

        if ui == vi:
            if self.cppqm.vartype(ui) == cppVartype.SPIN:
                raise ValueError(f"SPIN variables (e.g. {self.variables[ui]!r}) "
                                 "cannot have interactions with themselves"
                                 )
            if self.cppqm.vartype(ui) == cppVartype.BINARY:
                raise ValueError(f"BINARY variables (e.g. {self.variables[ui]!r}) "
                                 "cannot have interactions with themselves"
                                 )

        if not self.REAL_INTERACTIONS:
            if self.cppqm.vartype(ui) == cppVartype.REAL:
                raise ValueError(
                    f"REAL variables (e.g. {self.variables[ui]!r}) "
                    "cannot have interactions"
                    )
            if self.cppqm.vartype(vi) == cppVartype.REAL:
                raise ValueError(
                    f"REAL variables (e.g. {self.variables[vi]!r}) "
                    "cannot have interactions"
                    )

        self.cppqm.add_quadratic(ui, vi, bias)

    def add_quadratic(self, object u, object v, bias_type bias):
        cdef Py_ssize_t ui = self.variables.index(u)
        cdef Py_ssize_t vi = self.variables.index(v)
        self._add_quadratic(ui, vi, bias)

    def add_quadratic_from_arrays(self,
                                  ConstInteger[::1] irow, ConstInteger[::1] icol,
                                  ConstNumeric[::1] qdata):
        if not irow.shape[0] == icol.shape[0] == qdata.shape[0]:
            raise ValueError("quadratic vectors should be equal length")
        cdef Py_ssize_t length = irow.shape[0]

        cdef Py_ssize_t vi
        if self.variables._is_range():
            if length > self.num_variables():
                raise ValueError("variables must already exist")

            for vi in range(length):
                self._add_quadratic(irow[vi], icol[vi], qdata[vi])
        else:
            for vi in range(length):
                self.add_quadratic(irow[vi], icol[vi], qdata[vi])

    def add_quadratic_from_iterable(self, quadratic):
        cdef Py_ssize_t ui, vi
        cdef bias_type bias
        for u, v, bias in quadratic:
            ui = self.variables.index(u)
            vi = self.variables.index(v)
            self._add_quadratic(ui, vi, bias)

    def add_variable(self, vartype, label=None, *, lower_bound=None, upper_bound=None):
        if not isinstance(vartype, Vartype):  # redundant, but provides a bit of a speedup
            vartype = as_vartype(vartype, extended=True)
        cdef cppVartype cppvartype = self.cppvartype(vartype)

        cdef bias_type lb
        cdef bias_type ub

        cdef Py_ssize_t vi
        if label is not None and self.variables.count(label):
            # it already exists, so check that vartype matches
            vi = self.variables.index(label)
            if self.cppqm.vartype(vi) != cppvartype:
                raise TypeError(f"variable {label!r} already exists with a different vartype")
            if cppvartype != cppVartype.BINARY and cppvartype != cppVartype.SPIN:
                if lower_bound is not None:
                    lb = lower_bound
                    if lb != self.cppqm.lower_bound(vi):
                        raise ValueError(
                            f"the specified lower bound, {lower_bound}, for "
                            f"variable {label!r} is different than the existing lower "
                            f"bound, {int(self.cppqm.lower_bound(vi))}")
                if upper_bound is not None:
                    ub = upper_bound
                    if ub != self.cppqm.upper_bound(vi):
                        raise ValueError(
                            f"the specified upper bound, {upper_bound}, for "
                            f"variable {label!r} is different than the existing upper "
                            f"bound, {int(self.cppqm.upper_bound(vi))}")

            return label

        if cppvartype == cppVartype.BINARY or cppvartype == cppVartype.SPIN:
            # in this case we just ignore the provided values
            lb = cppvartype_info[bias_type].default_min(cppvartype)
            ub = cppvartype_info[bias_type].default_max(cppvartype)
        elif cppvartype == cppVartype.INTEGER or cppvartype == cppVartype.REAL:
            if lower_bound is None:
                lb = cppvartype_info[bias_type].default_min(cppvartype)
            else:
                lb = lower_bound
                if lb < cppvartype_info[bias_type].min(cppvartype):
                    raise ValueError(f"lower_bound cannot be less than {cppvartype_info[bias_type].min(cppvartype)}")

            if upper_bound is None:
                ub = cppvartype_info[bias_type].default_max(cppvartype)
            else:
                ub = upper_bound
                if ub > cppvartype_info[bias_type].max(cppvartype):
                    raise ValueError(f"upper_bound cannot be greater than {cppvartype_info[bias_type].max(cppvartype)}")
            
            if lb > ub:
                raise ValueError("lower_bound must be less than or equal to upper_bound")

            if cppvartype == cppVartype.INTEGER and ceil(lb) > floor(ub):
                raise ValueError("there must be at least one valid integer between lower_bound and upper_bound")
        else:
            raise RuntimeError("unknown vartype")

        self.cppqm.add_variable(cppvartype, lb, ub)

        self.variables._append(label)

        assert self.cppqm.num_variables() == self.variables.size()

        return self.variables.at(-1)

    def change_vartype(self, vartype, v):
        vartype = as_vartype(vartype, extended=True)
        cdef Py_ssize_t vi = self.variables.index(v)
        try:
            self.cppqm.change_vartype(self.cppvartype(vartype), vi)
        except RuntimeError:
            # c++ logic_error
            raise TypeError(f"cannot change vartype {self.vartype(v).name!r} "
                            f"to {vartype.name!r}") from None

    cdef cppVartype cppvartype(self, object vartype) except? cppVartype.SPIN:
        return cppvartype(vartype)

    cdef const cppQuadraticModel[bias_type, index_type]* data(self):
        """Return a pointer to the C++ QuadraticModel."""
        return self.cppqm

    @classmethod
    def from_cybqm(cls, cyBQM bqm):

        cdef cyQM_template qm = cls()

        qm.offset = bqm.offset

        # linear
        cdef Py_ssize_t vi
        cdef cppVartype vartype = bqm.cppbqm.vartype()
        for vi in range(bqm.num_variables()):
            qm.cppqm.add_variable(vartype)
            qm.cppqm.set_linear(vi, bqm.cppbqm.linear(vi))
        qm.variables._extend(bqm.variables)

        # quadratic
        it = bqm.cppbqm.cbegin_quadratic()
        while it != bqm.cppbqm.cend_quadratic():
            qm.cppqm.set_quadratic(deref(it).u, deref(it).v, deref(it).bias)
            inc(it)

        return qm

    def set_linear(self, v, bias_type bias):
        cdef Py_ssize_t vi = self.variables.index(v)
        self.cppqm.set_linear(vi, bias)

    def set_lower_bound(self, v, bias_type lb):
        cdef Py_ssize_t vi = self.variables.index(v)
        cdef cppVartype cppvartype = self.cppqm.vartype(vi)

        if cppvartype == cppVartype.BINARY or cppvartype == cppVartype.SPIN:
            raise ValueError(
                "cannot set the lower bound for BINARY or SPIN variables, "
                f"{v!r} is a {self.vartype(v).name} variable")

        if lb < cppvartype_info[bias_type].min(cppvartype):
            raise ValueError(f"lower_bound cannot be less than {cppvartype_info[bias_type].min(cppvartype)}")
            
        if lb > self.cppqm.upper_bound(vi):
            raise ValueError(
                f"the specified lower bound, {lb}, cannot be set greater than the "
                f"current upper bound, {self.cppqm.upper_bound(vi)}"
                )

        if cppvartype == cppVartype.INTEGER:
            if ceil(lb) > floor(self.cppqm.upper_bound(vi)):
                raise ValueError(
                    "there must be at least one integer value between "
                    f"the specified lower bound, {lb} and the "
                    f"current upper bound, {self.cppqm.upper_bound(vi)}"
                    )

        self.cppqm.set_lower_bound(vi, lb)

    def set_upper_bound(self, v, bias_type ub):
        cdef Py_ssize_t vi = self.variables.index(v)
        cdef cppVartype cppvartype = self.cppqm.vartype(vi)

        if cppvartype == cppVartype.BINARY or cppvartype == cppVartype.SPIN:
            raise ValueError(
                "cannot set the upper bound for BINARY or SPIN variables, "
                f"{v!r} is a {self.vartype(v).name} variable")

        if ub > cppvartype_info[bias_type].max(cppvartype):
            raise ValueError(f"upper_bound cannot be more than {cppvartype_info[bias_type].max(cppvartype)}")
            
        if ub < self.cppqm.lower_bound(vi):
            raise ValueError(
                f"the specified upper bound, {ub}, cannot be set less than the "
                f"current lower bound, {self.cppqm.lower_bound(vi)}"
                )

        if cppvartype == cppVartype.INTEGER:
            if ceil(self.cppqm.lower_bound(vi)) > floor(ub):
                raise ValueError(
                    "there must be at least one integer value between "
                    f"the specified upper bound, {ub} and the "
                    f"current lower bound, {self.cppqm.lower_bound(vi)}"
                    )

        self.cppqm.set_upper_bound(vi, ub)

    def set_quadratic(self, u, v, bias_type bias):
        cdef Py_ssize_t ui = self.variables.index(u)
        cdef Py_ssize_t vi = self.variables.index(v)

        if ui == vi:
            if self.cppqm.vartype(ui) == cppVartype.SPIN:
                raise ValueError(f"SPIN variables (e.g. {self.variables[ui]!r}) "
                                 "cannot have interactions with themselves"
                                 )
            if self.cppqm.vartype(ui) == cppVartype.BINARY:
                raise ValueError(f"BINARY variables (e.g. {self.variables[ui]!r}) "
                                 "cannot have interactions with themselves"
                                 )

        if not self.REAL_INTERACTIONS:
            if self.cppqm.vartype(ui) == cppVartype.REAL:
                raise ValueError(
                    f"REAL variables (e.g. {self.variables[ui]!r}) "
                    "cannot have interactions"
                    )
            if self.cppqm.vartype(vi) == cppVartype.REAL:
                raise ValueError(
                    f"REAL variables (e.g. {self.variables[vi]!r}) "
                    "cannot have interactions"
                    )

        self.cppqm.set_quadratic(ui, vi, bias)

    def update(self, cyBQM_and_QM other):
        # we'll need a mapping from the other's variables to ours
        cdef vector[Py_ssize_t] mapping
        mapping.reserve(other.num_variables())

        cdef Py_ssize_t vi

        # first make sure that any variables that overlap match in terms of
        # vartype and bounds
        for vi in range(other.num_variables()):
            v = other.variables.at(vi)
            if self.variables.count(v):
                # there is a variable already
                mapping.push_back(self.variables.index(v))

                if self.cppqm.vartype(mapping[vi]) != other.data().vartype(vi):
                    raise ValueError(f"conflicting vartypes: {v!r}")

                if self.cppqm.lower_bound(mapping[vi]) != other.data().lower_bound(vi):
                    raise ValueError(f"conflicting lower bounds: {v!r}")

                if self.cppqm.upper_bound(mapping[vi]) != other.data().upper_bound(vi):
                    raise ValueError(f"conflicting upper bounds: {v!r}")
            else:
                # not yet present, let's just track that fact for now
                # in case there is a mismatch so we don't modify our object yet
                mapping.push_back(-1)

        for vi in range(mapping.size()):
            if mapping[vi] != -1:
                continue  # already added and checked

            mapping[vi] = self.num_variables()  # we're about to add a new one

            v = other.variables.at(vi)
            vartype = other.vartype(v)

            self.add_variable(vartype, v,
                              lower_bound=other.data().lower_bound(vi),
                              upper_bound=other.data().upper_bound(vi),
                              )

        # variables are in place!
        
        # the linear biases
        for vi in range(mapping.size()):
            self.cppqm.add_linear(mapping[vi], other.data().linear(vi))

        # the quadratic biases
        # dev note: for even more speed we could check that mapping is
        # a range, and in that case can just add them without the indirection
        # or the sorting.
        it = other.data().cbegin_quadratic()
        while it != other.data().cend_quadratic():
            self.cppqm.add_quadratic(
                mapping[deref(it).u],
                mapping[deref(it).v],
                deref(it).bias
                )
            inc(it)

        # the offset
        self.cppqm.add_offset(other.data().offset())



# distutils: include_dirs = dimod/include/

# Copyright 2022 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from libcpp.utility cimport pair
from dimod.libcpp.vartypes cimport Vartype

__all__ = ['BinaryQuadraticModelBase']

cdef extern from "dimod/abc.h" namespace "dimod::abc" nogil:
    cdef cppclass QuadraticModelBase[Bias, Index]:

        ctypedef Bias bias_type
        ctypedef Index index_type
        ctypedef size_t size_type

        cppclass const_neighborhood_iterator:
            cppclass value_type:
                index_type v
                bias_type bias

                # todo: change to v, bias appropriately
                index_type first "v"
                bias_type second "bias"

            const_neighborhood_iterator() except +
            const_neighborhood_iterator(const_neighborhood_iterator&) except +
            operator=(const_neighborhood_iterator&) except +
            const value_type& operator*()
            const_neighborhood_iterator operator++()
            const_neighborhood_iterator operator--()
            const_neighborhood_iterator operator++(int)
            const_neighborhood_iterator operator--(int)
            const_neighborhood_iterator operator+(size_type)
            const_neighborhood_iterator operator-(size_type)
            ptrdiff_t operator-(const_neighborhood_iterator)
            bint operator==(const_neighborhood_iterator)
            bint operator!=(const_neighborhood_iterator)
            bint operator<(const_neighborhood_iterator)
            bint operator>(const_neighborhood_iterator)
            bint operator<=(const_neighborhood_iterator)
            bint operator>=(const_neighborhood_iterator)


        cppclass const_quadratic_iterator:
            cppclass value_type:
                index_type u
                index_type v
                bias_type bias

            const_quadratic_iterator()
            const_quadratic_iterator(const_quadratic_iterator&) except +
            operator=(const_quadratic_iterator&) except +
            const value_type& operator*()
            const_quadratic_iterator operator++()
            const_quadratic_iterator operator++(int)
            bint operator==(const_quadratic_iterator)
            bint operator!=(const_quadratic_iterator)

        # developer note: we avoid any overloads, due to a bug in Cython
        # https://github.com/cython/cython/issues/1357
        # https://github.com/cython/cython/issues/1868

        void add_linear(index_type, bias_type)
        void add_offset(bias_type)
        void add_quadratic(index_type, index_type, bias_type)
        void add_quadratic_from_coo "add_quadratic" [ItRow, ItCol, ItBias](ItRow, ItCol, ItBias, index_type)
        void add_quadratic_back(index_type, index_type, bias_type)
        void add_quadratic_from_dense[T](const T dense[], index_type)
        const_neighborhood_iterator cbegin_neighborhood(index_type)
        const_neighborhood_iterator cend_neighborhood(index_type)
        const_quadratic_iterator cbegin_quadratic()
        const_quadratic_iterator cend_quadratic()
        void clear()
        bias_type energy[Iter](Iter)
        void fix_variable[T](index_type, T)
        bint is_linear()
        bias_type linear(index_type)
        bias_type lower_bound(index_type)
        size_type nbytes()
        size_type nbytes(bint)
        size_type num_interactions()
        size_type degree "num_interactions" (index_type)
        size_type num_variables()
        bias_type offset()
        bias_type quadratic(index_type, index_type)
        bias_type quadratic_at(index_type, index_type) except+
        bint remove_interaction(index_type, index_type)
        void remove_variable(index_type)
        void scale(bias_type)
        void set_linear(index_type, bias_type)
        void set_offset(bias_type)
        void set_quadratic(index_type, index_type, bias_type) except+
        bias_type upper_bound(index_type)
        Vartype vartype(index_type)



# Copyright 2021 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import operator

cimport cython

from cython.operator cimport preincrement as inc, dereference as deref
from libcpp.algorithm cimport lower_bound as cpplower_bound

from dimod.libcpp.vartypes cimport Vartype as cppVartype

from dimod.cyutilities cimport as_numpy_float
from dimod.cyutilities cimport ConstNumeric
from dimod.sampleset import as_samples
from dimod.variables import Variables
from dimod.vartypes import Vartype

# super minor performance benefit from constructing these here rather than
# in the __init__
_dtype = np.dtype(BIAS_DTYPE)
_index_dtype = np.dtype(INDEX_DTYPE)


cdef class cyQMBase_template:
    def __cinit__(self):
        # Dev note: we do *not* allocate self.base because it's an
        # abstract virtual class. So we have to rely on subclasses to do so

        self.variables = Variables()

        # These should be treated as immutable class attributes, but Cython
        # doesn't support them
        self.dtype = _dtype
        self.index_dtype = _index_dtype

    def __init__(self):
        # Ensure that this class is not instantiated by itself.
        # It is possible to get segfaults by bypassing this __init__, but that seems
        # like a small enough edge case to be OK.
        if self.base is NULL:
            raise TypeError(f"Can't instantiate abstract class {type(self).__name__}")

    @property
    def offset(self):
        return as_numpy_float(self.base.offset())

    @offset.setter
    def offset(self, bias_type offset):
        self.base.set_offset(offset)

    def clear(self):
        self.base.clear()
        self.variables._clear()

    def degree(self, v):
        cdef Py_ssize_t vi = self.variables.index(v)
        return self.base.degree(vi)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def _energies(self, ConstNumeric[:, ::1] samples, cyVariables labels):
        cdef Py_ssize_t num_samples = samples.shape[0]
        cdef Py_ssize_t num_variables = samples.shape[1]

        if num_variables != labels.size():
            # as_samples should never return inconsistent sizes, but we do this
            # check because the boundscheck is off and we otherwise might get
            # segfaults
            raise RuntimeError("as_samples returned an inconsistent samples/variables")

        # get the indices of the QM variables
        cdef Py_ssize_t[::1] qm_to_sample = np.empty(self.num_variables(), dtype=np.intp)
        cdef Py_ssize_t si
        for si in range(self.num_variables()):
            qm_to_sample[si] = labels.index(self.variables.at(si))

        cdef np.float64_t[::1] energies = np.empty(num_samples, dtype=np.float64)

        # alright, now let's calculate some energies!
        cdef Py_ssize_t ui, vi
        for si in range(num_samples):
            # offset
            energies[si] = self.base.offset()

            for ui in range(self.num_variables()):
                # linear
                energies[si] += self.base.linear(ui) * samples[si, qm_to_sample[ui]];

                it = self.base.cbegin_neighborhood(ui)
                end = self.base.cend_neighborhood(ui)
                while it != end and deref(it).v <= ui:
                    vi = deref(it).v

                    energies[si] += deref(it).bias * samples[si, qm_to_sample[ui]] * samples[si, qm_to_sample[vi]]

                    inc(it)

        return energies

    def energies(self, samples_like, dtype=None):
        # todo: deprecate dtype, it doesn't actually change what dtype
        # the calculation is done in, so it's pretty misleading

        samples, labels = as_samples(samples_like, labels_type=Variables)

        # we need contiguous and unsigned. as_samples actually enforces contiguous
        # but no harm in double checking for some future-proofness
        samples = np.ascontiguousarray(
                samples,
                dtype=f'i{samples.dtype.itemsize}' if np.issubdtype(samples.dtype, np.unsignedinteger) else None,
                )

        try:
            return np.asarray(self._energies(samples, labels), dtype=dtype)
        except TypeError as err:
            if np.issubdtype(samples.dtype, np.floating) or np.issubdtype(samples.dtype, np.signedinteger):
                raise err
            raise ValueError(f"unsupported sample dtype: {samples.dtype.name}")

    def get_linear(self, v):
        return as_numpy_float(self.base.linear(self.variables.index(v)))

    def get_quadratic(self, u, v, default=None):
        cdef Py_ssize_t ui = self.variables.index(u)
        cdef Py_ssize_t vi = self.variables.index(v)

        if ui == vi:
            if self.base.vartype(ui) == cppVartype.SPIN:
                raise ValueError(f"SPIN variables (e.g. {u!r}) "
                                 "cannot have interactions with themselves"
                                 )
            if self.base.vartype(ui) == cppVartype.BINARY:
                raise ValueError(f"BINARY variables (e.g. {v!r}) "
                                 "cannot have interactions with themselves"
                                 )
        cdef bias_type bias
        try:
            bias = self.base.quadratic_at(ui, vi)
        except IndexError:
            # out of range error is automatically converted to IndexError
            if default is None:
                raise ValueError(f"{u!r} and {v!r} have no interaction") from None
            bias = default
        return as_numpy_float(bias)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def _ilinear(self):
        """Return the linear biases in a numpy array."""
        cdef bias_type[:] ldata = np.empty(self.num_variables(), dtype=self.dtype)
        cdef Py_ssize_t vi
        for vi in range(self.num_variables()):
            ldata[vi] = self.base.linear(vi)
        return ldata

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def _ineighborhood(self, Py_ssize_t vi, bint lower_triangle=False):
        """Return a NumPy structured array containing the neighborhood of vi."""
        if not 0 <= vi < self.num_variables():
            raise ValueError(f"out of range variable index: {vi}")

        cdef Py_ssize_t length  # length of array we wish to allocate
        if lower_triangle:
            begin = self.base.cbegin_neighborhood(vi)
            loc = cpplower_bound(begin, self.base.cend_neighborhood(vi), vi)
            
            # We want to include self-loops
            if loc != self.base.cend_neighborhood(vi) and deref(loc).v == vi:
                inc(loc)

            length = loc - begin
        else:
            length = self.base.degree(vi)

        # Make a NumPy struct array. 
        # We choose our labels to match dimod::abc::OneVarTerm
        # We pack the values by setting align=False
        neighborhood = np.empty(
            length, dtype=np.dtype([('v', self.index_dtype), ('bias', self.dtype)], align=False))

        cdef index_type[:] index_view = neighborhood['v']
        cdef bias_type[:] biases_view = neighborhood['bias']

        # now walk through the neighborhood
        it = self.base.cbegin_neighborhood(vi)
        for i in range(length):
            index_view[i] = deref(it).v
            biases_view[i] = deref(it).bias
            inc(it)
        
        return neighborhood

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def _ivarinfo(self):
        cdef Py_ssize_t num_variables = self.num_variables()

        # we could use the bias_type size to determine the vartype dtype to get
        # more alignment, but it complicates the code, so let's keep it simple
        # We choose the field names to mirror the internals of the QuadraticModel
        dtype = np.dtype([('vartype', np.int8), ('lb', self.dtype), ('ub', self.dtype)],
                         align=False)
        varinfo = np.empty(num_variables, dtype)

        cdef np.int8_t[:] vartype_view = varinfo['vartype']
        cdef bias_type[:] lb_view = varinfo['lb']
        cdef bias_type[:] ub_view = varinfo['ub']

        cdef Py_ssize_t vi
        for vi in range(self.num_variables()):
            vartype_view[vi] = self.base.vartype(vi)
            lb_view[vi] = self.base.lower_bound(vi)
            ub_view[vi] = self.base.upper_bound(vi)

        return varinfo

    cpdef bint is_linear(self):
        return self.base.is_linear()

    def iter_neighborhood(self, v):
        cdef Py_ssize_t vi = self.variables.index(v)

        it = self.base.cbegin_neighborhood(vi)
        while it != self.base.cend_neighborhood(vi):
            yield self.variables.at(deref(it).v), as_numpy_float(deref(it).bias)
            inc(it)

    def iter_quadratic(self):
        it = self.base.cbegin_quadratic()
        while it != self.base.cend_quadratic():
            u = self.variables.at(deref(it).u)
            v = self.variables.at(deref(it).v)
            yield u, v, as_numpy_float(deref(it).bias)
            inc(it)

    def lower_bound(self, v):
        cdef Py_ssize_t vi = self.variables.index(v)
        cdef bias_type lb = self.base.lower_bound(vi)
        return as_numpy_float(lb)

    def nbytes(self, bint capacity = False):
        return self.base.nbytes(capacity)

    cpdef Py_ssize_t num_interactions(self):
        return self.base.num_interactions()

    cpdef Py_ssize_t num_variables(self):
        return self.base.num_variables()

    def reduce_linear(self, function, initializer=None):
        if self.num_variables() == 0 and initializer is None:
            # feels like this should be a ValueError but python raises
            # TypeError so...
            raise TypeError("reduce_linear() on an empty model")

        cdef Py_ssize_t start, vi
        cdef bias_type value, tmp

        if initializer is None:
            start = 1
            value = self.base.linear(0)
        else:
            start = 0
            value = initializer

        # speed up a few common cases
        if function is operator.add:
            for vi in range(start, self.num_variables()):
                value += self.base.linear(vi)
        elif function is max:
            for vi in range(start, self.num_variables()):
                tmp = self.base.linear(vi)
                if tmp > value:
                    value = tmp
        elif function is min:
            for vi in range(start, self.num_variables()):
                tmp = self.base.linear(vi)
                if tmp < value:
                    value = tmp
        else:
            for vi in range(start, self.num_variables()):
                value = function(value, self.base.linear(vi))

        return as_numpy_float(value)

    def reduce_neighborhood(self, u, function, initializer=None):
        cdef Py_ssize_t ui = self.variables.index(u)

        if self.base.degree(ui) == 0 and initializer is None:
            # feels like this should be a ValueError but python raises
            # TypeError so...
            raise TypeError("reduce_neighborhood() on an empty neighbhorhood")

        cdef bias_type value, tmp

        # span = self.base.neighborhood(ui)
        it = self.base.cbegin_neighborhood(ui)

        if initializer is None:
            value = deref(it).bias
            inc(it)
        else:
            value = initializer

        # speed up a few common cases
        if function is operator.add:
            while it != self.base.cend_neighborhood(ui):
                value += deref(it).bias
                inc(it)
        elif function is max:
            while it != self.base.cend_neighborhood(ui):
                tmp = deref(it).bias
                if tmp > value:
                    value = tmp
                inc(it)
        elif function is min:
            while it != self.base.cend_neighborhood(ui):
                tmp = deref(it).bias
                if tmp < value:
                    value = tmp
                inc(it)
        else:
            while it != self.base.cend_neighborhood(ui):
                value = function(value, deref(it).second)
                inc(it)

        return as_numpy_float(value)

    def reduce_quadratic(self, function, initializer=None):

        if self.base.is_linear() and initializer is None:
            # feels like this should be a ValueError but python raises
            # TypeError so...
            raise TypeError("reduce_quadratic() on a linear model")

        cdef bias_type value, tmp

        start = self.base.cbegin_quadratic()

        if initializer is None:
            value = deref(start).bias
            inc(start)
        else:
            value = initializer

        # handle a few common cases
        if function is operator.add:
            while start != self.base.cend_quadratic():
                value += deref(start).bias
                inc(start)
        elif function is max:
            while start != self.base.cend_quadratic():
                tmp = deref(start).bias
                if tmp > value:
                    value = tmp
                inc(start)
        elif function is min:
            while start != self.base.cend_quadratic():
                tmp = deref(start).bias
                if tmp < value:
                    value = tmp
                inc(start)
        else:
            while start != self.base.cend_quadratic():
                value = function(value, deref(start).bias)
                inc(start)

        return as_numpy_float(value)

    def relabel_variables(self, mapping):
        self.variables._relabel(mapping)

    def relabel_variables_as_integers(self):
        return self.variables._relabel_as_integers()

    def remove_interaction(self, u, v):
        cdef Py_ssize_t ui = self.variables.index(u)
        cdef Py_ssize_t vi = self.variables.index(v)
        
        if not self.base.remove_interaction(ui, vi):
            raise ValueError(f"{u!r} and {v!r} have no interaction")

    def remove_variable(self, v=None):
        if v is None:
            try:
                v = self.variables[-1]
            except IndexError:
                raise ValueError("cannot pop from an empty model")

        self.base.remove_variable(self.variables.index(v))
        self.variables._remove(v)

        return v

    def scale(self, bias_type scalar):
        self.base.scale(scalar)

    def upper_bound(self, v):
        cdef Py_ssize_t vi = self.variables.index(v)
        return as_numpy_float(self.base.upper_bound(vi))

    def vartype(self, v):
        cdef Py_ssize_t vi = self.variables.index(v)
        cdef cppVartype cppvartype = self.base.vartype(vi)

        if cppvartype == cppVartype.BINARY:
            return Vartype.BINARY
        elif cppvartype == cppVartype.SPIN:
            return Vartype.SPIN
        elif cppvartype == cppVartype.INTEGER:
            return Vartype.INTEGER
        elif cppvartype == cppVartype.REAL:
            return Vartype.REAL
        else:
            raise RuntimeError("unexpected vartype")



# distutils: language = c++
# cython: language_level=3

# Copyright 2021 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

cimport numpy as np

from dimod.cyqmbase.cyqmbase_float64 cimport cyQMBase_float64 as cyQMBase, bias_type, index_type

__all__ = ['cyBQM_float64']

include "cybqm_template.pxd.pxi"

cdef class cyBQM_float64(cyBQM_template):
    pass



# distutils: include_dirs = dimod/include/

# Copyright 2022 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from libcpp.memory cimport weak_ptr
from libcpp.vector cimport vector

from dimod.libcpp.abc cimport QuadraticModelBase
from dimod.libcpp.vartypes cimport Vartype

__all__ = ['ConstrainedQuadraticModel', 'Sense', 'Penalty', 'Constraint', 'Expression']


cdef extern from "dimod/constrained_quadratic_model.h" namespace "dimod" nogil:
    enum Sense:
        EQ
        LE
        GE

    enum Penalty:
        LINEAR
        QUADRATIC
        CONSTANT

    cdef cppclass Expression[Bias, Index](QuadraticModelBase[Bias, Index]):
        ctypedef Bias bias_type
        ctypedef Index index_type
        ctypedef size_t size_type

        cppclass const_quadratic_iterator:
            cppclass value_type:
                index_type u
                index_type v
                bias_type bias

            const_quadratic_iterator()
            const_quadratic_iterator(const_quadratic_iterator&) except +
            operator=(const_quadratic_iterator&) except +
            const value_type& operator*()
            const_quadratic_iterator operator++()
            const_quadratic_iterator operator++(int)
            bint operator==(const_quadratic_iterator)
            bint operator!=(const_quadratic_iterator)

        cppclass const_neighborhood_iterator:
            cppclass value_type:
                index_type v
                bias_type bias

            const_neighborhood_iterator()
            const_neighborhood_iterator(const_neighborhood_iterator&) except +
            operator=(const_neighborhood_iterator&) except +
            const value_type& operator*()
            const_neighborhood_iterator operator++()
            const_neighborhood_iterator operator++(int)
            bint operator==(const_neighborhood_iterator)
            bint operator!=(const_neighborhood_iterator)


        # void add_offset(bias_type)
        # void add_linear(index_type, bias_type)
        # void add_quadratic(index_type, index_type, bias_type)
        # void clear()
        # size_type num_variables()

        bint has_variable(index_type)
        bint is_disjoint(const Expression&)

        # exprconst_quadratic_iterator cbegin_quadratic()
        # exprconst_quadratic_iterator cend_quadratic()

        vector[index_type]& variables()

    cdef cppclass Constraint[Bias, Index](Expression[Bias, Index]):
        ctypedef Bias bias_type
        ctypedef Index index_type
        ctypedef size_t size_type

        bias_type rhs()
        Sense sense()
        Penalty penalty()
        bias_type weight()

        bint is_onehot()
        bint is_soft()
        void mark_discrete();
        void mark_discrete(bint mark);
        bint marked_discrete() const;
        void set_rhs(bias_type)
        void set_sense(Sense)
        void set_penalty(Penalty)
        void set_weight(bias_type)


    cdef cppclass ConstrainedQuadraticModel[bias_type, index_type]:
        Expression[bias_type, index_type] objective
        # vector[Constraint[bias_type, index_type]] constraints

        index_type add_constraint()
        index_type add_constraint(Constraint[bias_type, index_type]) except+
        index_type add_constraint[B, I, T](QuadraticModelBase[B, I]&, Sense, bias_type, vector[T])
        index_type add_constraints(index_type)
        index_type add_variable(Vartype)
        index_type add_variable(Vartype, bias_type, bias_type)
        void change_vartype(Vartype, index_type) except+
        void clear()
        Constraint[bias_type, index_type]& constraint_ref(index_type)
        weak_ptr[Constraint[bias_type, index_type]] constraint_weak_ptr(index_type)
        void fix_variable[T](index_type, T)
        bias_type lower_bound(index_type)
        Constraint[bias_type, index_type] new_constraint()
        size_t num_constraints()
        size_t num_interactions()
        size_t num_variables()
        void remove_constraint(index_type)
        void remove_variable(index_type)
        void set_lower_bound(index_type, bias_type)
        void set_objective[B, I](QuadraticModelBase&)
        void set_objective[B, I, T](QuadraticModelBase[B, I]&, vector[T])
        void set_upper_bound(index_type, bias_type)
        void substitute_variable(index_type, bias_type, bias_type)
        bias_type upper_bound(index_type)
        Vartype vartype(index_type)



# Copyright 2020 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""A sampler that returns the provided initial states."""
import typing

from dimod.binary.binary_quadratic_model import BinaryQuadraticModel
from dimod.core import Sampler, Initialized
from dimod.sampleset import SampleSet

from dimod.typing import SamplesLike
try:
    InitialStateGenerator = typing.Literal['none', 'tile', 'random']
except AttributeError:
    InitialStateGenerator = str

__all__ = ['IdentitySampler']


class IdentitySampler(Sampler, Initialized):
    """A sampler that can return, expand, or truncate the provided initial states.

    Examples:

    >>> samples = [{'a': 1, 'b': 0}, {'a': 0, 'b': 1}]
    >>> Q = {('a', 'b'): -1}
    >>> sampler = dimod.IdentitySampler()
    >>> sampleset = sampler.sample_qubo(Q, initial_states=samples)
    >>> print(sampleset)
       a  b energy num_oc.
    0  1  0    0.0       1
    1  0  1    0.0       1
    ['BINARY', 2 rows, 2 samples, 2 variables]

    Add randomly generated samples:

    >>> sampleset = sampler.sample_qubo(Q, initial_states=samples, num_reads=4)
    >>> print(sampleset)                                  # doctest: +SKIP
       a  b energy num_oc.
    2  1  1   -1.0       1
    3  1  1   -1.0       1
    0  1  0    0.0       1
    1  0  1    0.0       1
    ['BINARY', 4 rows, 4 samples, 2 variables]

    Note that in the above output, the specified initial states are printed after
    the generated lower-energy samples but are the first samples in the
    :attr:`~dimod.SampleSet.record`.

    """

    parameters: dict = {'initial_states': [],
                  'initial_states_generator': [],
                  'num_reads': [],
                  'seed': [],
                  }
    """Keyword arguments accepted by the sampling methods.

    Contents are a dict with the following keys:
    ``{'initial_states': [], 'initial_states_generator': [], 'num_reads': [], 'seed': []}``
    """

    properties = {}

    def sample(self, bqm: BinaryQuadraticModel,
               initial_states: typing.Optional[SamplesLike] = None,
               initial_states_generator: InitialStateGenerator = 'random',
               num_reads: typing.Optional[int] = None,
               seed: typing.Optional[int] = None,
               **kwargs) -> SampleSet:
        """Return, expand, or truncate the provided initial states.

        Args:
            bqm: Binary quadratic model to be sampled.

            initial_states:
                One or more samples, each defining an initial state for all the
                problem variables. Initial states are given one per read, but
                if fewer than ``num_reads`` initial states are defined,
                additional values are generated as specified by
                ``initial_states_generator``. See :func:`.as_samples` for a
                description of `samples-like`.

            initial_states_generator:
                Defines the expansion of ``initial_states`` if fewer than
                ``num_reads`` are specified:

                * "none":
                    If the number of initial states specified is smaller than
                    ``num_reads``, raises ``ValueError``.

                * "tile":
                    Reuses the specified initial states if fewer than
                    ``num_reads``, or truncates if greater.

                * "random":
                    Expands the specified initial states with randomly
                    generated states if fewer than ``num_reads``, or truncates if
                    greater.

            num_reads:
                Number of reads. If not explicitly given, the number of reads is
                set to the number of initial states, or if initial states are
                not given, to the default of 1.

            seed:
                Seed (32-bit unsigned integer) to use for the PRNG. Specifying a
                particular seed with a constant set of parameters produces
                identical results. If not provided, a random seed is chosen.

        Returns:
            A :obj:`.SampleSet` with the specified initial states, optionally
            truncated or augmented.

        """
        kwargs = self.remove_unknown_kwargs(**kwargs)
        parsed = self.parse_initial_states(bqm, initial_states,
                   initial_states_generator, num_reads, seed, **kwargs)

        # and we're done, we just pass back the initial state we were handed
        return parsed.initial_states
